#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>
#include <sys/file.h>

#define SEEK_SET	0
#define SEEK_CUR	1

#ifndef O_ACCMODE
#define O_ACCMODE (O_RDONLY | O_WRONLY | O_RDWR)
#endif

extern char	*malloc();
extern void	free();

/* EXACT TYPES */
typedef char 		int8e_type;
typedef unsigned char 	uint8e_type;
typedef short 		int16e_type;
typedef unsigned short 	uint16e_type;
typedef int 		int32e_type;
typedef unsigned int 	uint32e_type;

/* CORRECT SIZE OR GREATER */
typedef char 		int8_type;
typedef unsigned char 	uint8_type;
typedef short 		int16_type;
typedef unsigned short 	uint16_type;
typedef int 		int32_type;
typedef unsigned int 	uint32_type;
#include "fopen-same.h"
