typedef union {
  bfd_vma integer;
  int voidval;
  char *name;
  int token;
  union etree_union *etree;
  struct sec *section;
  struct lang_output_section_statement_struct *output_section_statement;
  union  lang_statement_union **statement_ptr;
  int lineno;
  struct {
    FILE *file;
    char *name;
    unsigned int lineno;
  } state;

  
} YYSTYPE;
#define	INT	258
#define	NAME	259
#define	PLUSEQ	260
#define	MINUSEQ	261
#define	MULTEQ	262
#define	DIVEQ	263
#define	LSHIFTEQ	264
#define	RSHIFTEQ	265
#define	ANDEQ	266
#define	OREQ	267
#define	OROR	268
#define	ANDAND	269
#define	EQ	270
#define	NE	271
#define	LE	272
#define	GE	273
#define	LSHIFT	274
#define	RSHIFT	275
#define	UNARY	276
#define	END	277
#define	ALIGN_K	278
#define	BLOCK	279
#define	LONG	280
#define	SHORT	281
#define	BYTE	282
#define	SECTIONS	283
#define	SIZEOF_HEADERS	284
#define	OUTPUT_FORMAT	285
#define	FORCE_COMMON_ALLOCATION	286
#define	OUTPUT_ARCH	287
#define	INCLUDE	288
#define	MEMORY	289
#define	DEFSYMEND	290
#define	NOLOAD	291
#define	DSECT	292
#define	COPY	293
#define	INFO	294
#define	OVERLAY	295
#define	DEFINED	296
#define	TARGET_K	297
#define	SEARCH_DIR	298
#define	MAP	299
#define	ENTRY	300
#define	OPTION_e	301
#define	OPTION_c	302
#define	OPTION_noinhibit_exec	303
#define	OPTION_s	304
#define	OPTION_S	305
#define	OPTION_sort_common	306
#define	OPTION_EB	307
#define	OPTION_EL	308
#define	OPTION_G	309
#define	OPTION_Gval	310
#define	OPTION_format	311
#define	OPTION_F	312
#define	OPTION_u	313
#define	OPTION_Bstatic	314
#define	OPTION_N	315
#define	SIZEOF	316
#define	NEXT	317
#define	ADDR	318
#define	OPTION_d	319
#define	OPTION_dc	320
#define	OPTION_dp	321
#define	OPTION_x	322
#define	OPTION_X	323
#define	OPTION_defsym	324
#define	OPTION_v	325
#define	OPTION_V	326
#define	OPTION_M	327
#define	OPTION_t	328
#define	STARTUP	329
#define	HLL	330
#define	SYSLIB	331
#define	FLOAT	332
#define	NOFLOAT	333
#define	OPTION_Map	334
#define	OPTION_n	335
#define	OPTION_r	336
#define	OPTION_o	337
#define	OPTION_b	338
#define	OPTION_R	339
#define	OPTION_relax	340
#define	OPTION_l	341
#define	OPTION_L	342
#define	OPTION_T	343
#define	OPTION_Aarch	344
#define	OPTION_Tfile	345
#define	OPTION_Texp	346
#define	OPTION_y	347
#define	OPTION_Ur	348
#define	ORIGIN	349
#define	FILL	350
#define	OPTION_g	351
#define	LENGTH	352
#define	CREATE_OBJECT_SYMBOLS	353
#define	INPUT	354
#define	OUTPUT	355
#define	CONSTRUCTORS	356
#define	OPTION_RETAIN_SYMBOLS_FILE	357
#define	ALIGNMOD	358
#define	AT	359
#define	CHIP	360
#define	LIST	361
#define	SECT	362
#define	ABSOLUTE	363
#define	LOAD	364
#define	NEWLINE	365
#define	ENDWORD	366
#define	ORDER	367
#define	NAMEWORD	368
#define	FORMAT	369
#define	PUBLIC	370
#define	BASE	371
#define	ALIAS	372
#define	TRUNCATE	373


extern YYSTYPE yylval;
